/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

/*
 * ColumnSpec.java
 *
 * Created on 11 janvier 2001, 16:57
 */

package org.xquark.repository.dbms;

import java.sql.Types;

import org.xquark.repository.util.Functions;

/**
 * Class containing information for relational column creation.
 *
 */
public class ColumnSpec
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name:  $";
    
    /* Column name */
    protected String name;
    protected int JDBCType;
    protected long size = -1;
    protected int prec = -1;
    protected int scale = -1;
    protected short paramSize = -1;
    protected short paramPrec = -1;
    protected short paramScale = -1;
    
    /* Is part of the primary key */
    protected boolean key = false;
    protected boolean nullable = true;

    //////////////////////////////////////////////////////////////////////////
    //                          CONSTRUCTORS
    //////////////////////////////////////////////////////////////////////////
    public ColumnSpec() {}
    
    /** Creates new ColumnSpec 
     * @param name The column name.
     * @param key true if the column belongs to the primary key.
     */
    public ColumnSpec(String name, int JDBCType, long size)
    {
        set(name, JDBCType, size);
    }
    
    /** Creates new ColumnSpec 
     * @param name The column name.
     * @param key true if the column belongs to the primary key.
     */
    public ColumnSpec(String name, int JDBCType, int prec, int scale)
    {
        set(name, JDBCType, prec, scale);
    }
    
    /** Creates new ColumnSpec 
     * @param name The column name.
     * @param key true if the column belongs to the primary key.
     */
    public ColumnSpec(String name, int JDBCType, boolean key)
    {
        set(name, JDBCType, key);
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                              ACCESSORS
    //////////////////////////////////////////////////////////////////////////
    
    public void setName(String name)
    {
        this.name = name;
    }
    
    public void set(String name, int JDBCType, long size)
    {
        this.name = name;
        this.JDBCType = JDBCType;
        this.size = size;
    }
    
    public void set(String name, int JDBCType, int prec, int scale)
    {
        set(name, JDBCType, size);
        this.prec = prec;
        this.scale = scale;
    }
    
    /** Accessor.
     * <p>Used for object recycling.</p>
     * @param name The column name.
     * @param key true if the column belongs to the primary key.
     */
    public void set(String name, int JDBCType, boolean key)
    {
        this.name = name;
        this.JDBCType = JDBCType;
        setKeyRole(key);
    }
    
    public void setDataType(String type)
    {
        JDBCType = Functions.parseJDBCType(type);
    }
    
    /** Accessor.
     * @param key true if the column belongs to the primary key.
     */
    public void setKeyRole(boolean key)
    {
        this.key = key;
    }
    
    public void setSize(long size, short param)
    {
        this.size = size;
        paramSize = param;
    }
    
    public void setPrecision(int prec, short param)
    {
        this.prec = prec;
        paramPrec = param;
    }
    
    public void setScale(int scale, short param)
    {
        this.scale = scale;
        paramScale = param;
    }
    
    public void setNullable(boolean nullable)
    {
        this.nullable = nullable;
    }
    
    /** Return the column name */
    public String getName()
    {
        return name;
    }
    
    public int getDataType()
    {
        return JDBCType;
    }
    
    /** Return true if the columns belongs to the primary key */
    public boolean isPrimaryKey()
    {
        return key;
    }
    
    public boolean isNullable()
    {
        return nullable;
    }
    
    public long getSize(long[] sizeParameters)
    {
        if (paramSize >= 0 && sizeParameters[paramSize] >= 0)
            return sizeParameters[paramSize];
        else
            return size;
    }
    
    public int getPrecision(long[] sizeParameters)
    {
        if (paramPrec >= 0 && sizeParameters[paramPrec] >= 0)
            return (int)sizeParameters[paramPrec];
        else
            return prec;
    }
    
    public int getScale(long[] sizeParameters)
    {
        if (paramScale >= 0 && sizeParameters[paramScale] >= 0)
            return (int)sizeParameters[paramScale];
        else
            return scale;
    }
    
    /**
     * Use default values (no parameter).
     */
    public String getCreationString(TypeMap typeMap)
    {
        return name + ' ' + getTypeCreationString(JDBCType, typeMap, size, prec,scale);
    }
    
    /**
     * Use default values (no parameter).
     */
    public String getTypeCreationString(TypeMap typeMap)
    {
        return getTypeCreationString(JDBCType, typeMap, size, prec,scale);
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                          PUBLIC STATIC UTILITIES
    //////////////////////////////////////////////////////////////////////////
    /* Generate the native SQL statement for column type specification. The 
     * optimum native type is chosen according to size parameters.
     */
    public static String getTypeDDL(int dataType, AbstractConnection dbmsInfo,
            long maxLength)
    {
        return getTypeDDLStatement(dataType, dbmsInfo, maxLength, -1, -1);
    }
    
    public static String getTypeDDL(int dataType, AbstractConnection dbmsInfo,
            int precision, int scale)
    {
        return getTypeDDLStatement(dataType, dbmsInfo, -1, precision, scale);
    }
    
    public static String getTypeDDL(int dataType, AbstractConnection dbmsInfo,
            int precision)
    {
        return getTypeDDLStatement(dataType, dbmsInfo, -1, precision, -1);
    }
    
    public static String getTypeDDL(int dataType, AbstractConnection dbmsInfo)
    {
        return getTypeDDLStatement(dataType, dbmsInfo, -1, -1, -1);
    }
    
    /* Warning: no size adaptation is performed (unlike getTypeDDL()).*/
    public static String getTypeCreationString(int dataType, TypeMap typeMap,
    long maxLength, int precision, int scale)
    {
        String typeString;
        
        /* building precision and scale string */
        switch (dataType) // switch according to target JDBC type for primitive type
        {   // and try to analyse facets(restrictions) to find a 'smaller one'
            case Types.CHAR : case Types.BINARY :
            case Types.VARCHAR : case Types.VARBINARY :
            case Types.DOUBLE : case Types.FLOAT :
            case Types.NUMERIC : case Types.DECIMAL :
                typeString = getDefaultTypeCreationString(dataType, typeMap,
                maxLength, precision, scale);
                break;
            default :   // for other type just concatenate the creation metadada param
                typeString = typeMap.getNativeTypeName(dataType);
                if (typeMap.getTypeParam(dataType) != null)
                    typeString += typeMap.getTypeParam(dataType);
        }
        return typeString;
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                          STATIC UTILITIES
    //////////////////////////////////////////////////////////////////////////
    static int findRealisticType(int type, long size, AbstractConnection dbmsInfo)
    {
        switch (type) // switch according to "base" target JDBC type
        {   // and try to analyse facets(restrictions) to find a 'smaller one'
            case Types.VARCHAR :
            case Types.LONGVARCHAR :
                if (size <= dbmsInfo.getTypeMap().getSize(Types.VARCHAR))
                    type = Types.VARCHAR;
                else
                    type = Types.LONGVARCHAR;
                break;
            case Types.VARBINARY :
            case Types.LONGVARBINARY :
                if (size <= dbmsInfo.getTypeMap().getSize(Types.VARBINARY))
                    type = Types.VARBINARY;
                else
                    type = Types.LONGVARBINARY;
                break;
            default :   // Nothing to do with VARCHAR, TIMESTAMP, REAL
        }
        return type;
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                              PRIVATE
    //////////////////////////////////////////////////////////////////////////
    /** Builds the database-specific string for type specification.
     * It may be constructed by concatenation of typeName, size, and scale
     * or in any other way.
     * <B>A call to the static method loadTypeMetadata must have been
     * performed before this method is able to return something.</B>
     * @return  a creation string assumed to be constructed
     * as follows : typeName + '(' + size + ',' + scale + ')'
     * null if loadTypeMetadata has not been called.
     */
    private static String getDefaultTypeCreationString(int dataType,
    TypeMap typeMap, long maxLength, int precision, int scale)
    {
        String ret = null;
        if (typeMap != null)
        {
            ret = typeMap.getNativeTypeName(dataType);
            if (maxLength > 0)
                ret += "(" + maxLength + ")";
            else if (precision > 0)
            {
                ret += "(" + precision;
                if (scale > 0)
                    ret += "," + scale;
                ret += ")";
            }
            else if (typeMap.getTypeParam(dataType) == null)    // take the maximum authorized by the rdbms
            {
                switch (dataType) // switch according to target JDBC type for primitive type
                {
                    case Types.NUMERIC : case Types.FLOAT :
                    case Types.DOUBLE : case Types.DECIMAL :
                        // do not add precision (max is implicit) // TO IMPROVE : build a flag for param needed
                        break;
                    default :   // for other type just concatenate the creation metadada param
                        ret += "(" + typeMap.getSize(dataType) + ")";
                }
            }
        }
        return ret;
    }
    
    private static String getTypeDDLStatement(int dataType, AbstractConnection dbmsInfo,
            long maxLength, int precision, int scale)
    {
        /* Adding maximum size for types needing it */
        if (maxLength < 0)
        {
            switch (dataType) // switch according to target JDBC type for primitive type
            {   // and try to analyse facets(restrictions) to find a 'smaller one'
                case Types.CHAR : case Types.VARCHAR :
                case Types.VARBINARY : case Types.BINARY :
                    maxLength = dbmsInfo.getTypeMap().getSize(dataType);
                    break;
                default :
            }
        }
        
       return getTypeCreationString(
            findRealisticType(dataType, maxLength, dbmsInfo),
            dbmsInfo.getTypeMap(),
            maxLength,
            precision,
            scale
            );
    }
}
